import hashlib
import cloudpickle


class Logger():
    def __init__(self, config):
        # check config.log.path


        # if os.path.exists(path) and os.path.isfile(path):
        #     self.file = open(path, 'a')
        # else:
        #     if not os.path.exists(os.path.dirname(path))
        #         os.makedirs(os.path.dirname(path))

        # assert os.path.isfile(path)

    def __exit__(self):


